from datetime import datetime

from sqlalchemy import create_engine
from sqlalchemy import DateTime
from sqlalchemy import Float
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import String
from sqlalchemy.engine import Engine
from sqlalchemy.schema import Column
from sqlalchemy.schema import Table


def generate_scorpion_testdb() -> Engine:
    """Create a test DB from Table 1 of the Scorpion paper (Wu & Madden,
    VLDB 2013).

    Rather than make sensor_id a foreign key, we make it a character
    type to test a wider variety of column types.
    """
    engine = create_engine('sqlite://')
    conn = engine.connect()
    metadata = MetaData()
    sensor_readings = Table(
        'sensor_readings', metadata,
        Column('id', Integer, primary_key=True),
        Column('created_at', DateTime, nullable=False),
        Column('sensor_id', String, nullable=False),
        Column('voltage', Float, nullable=False),
        Column('humidity', Float, nullable=False),
        Column('temperature', Float, nullable=False),
    )
    metadata.create_all(engine)
    values = [
        (datetime(2021, 5, 5, 11), '1', 2.64, 0.4, 34),
        (datetime(2021, 5, 5, 11), '2', 2.65, 0.5, 35),
        (datetime(2021, 5, 5, 11), '3', 2.63, 0.4, 35),
        (datetime(2021, 5, 5, 12), '1', 2.7, 0.3, 35),
        (datetime(2021, 5, 5, 12), '2', 2.7, 0.5, 35),
        (datetime(2021, 5, 5, 12), '3', 2.3, 0.4, 100),
        (datetime(2021, 5, 5, 13), '1', 2.7, 0.3, 35),
        (datetime(2021, 5, 5, 13), '2', 2.7, 0.5, 35),
        (datetime(2021, 5, 5, 13), '3', 2.3, 0.5, 80),
    ]
    conn.execute(sensor_readings.insert(), [
        dict(zip(
            ('created_at', 'sensor_id', 'voltage', 'humidity', 'temperature'),
            value))
        for value in values
    ])
    return engine


def generate_synthetic_testdb() -> Engine:
    """Create a synthetic database with several data types and
    distributions.
    For various data types (datetime, integer, string, float), we
    generate distributions that are:
    * Unique overall (every value is different)
    * The same overall (every value is the same)
    * Unique within a bucket (but repeats across buckets).
    """
    engine = create_engine('sqlite://')
    conn = engine.connect()
    metadata = MetaData()
    sensor_readings = Table(
        'synthetic_data', metadata,
        Column('id', Integer, primary_key=True),
        Column('same_datetime', DateTime, nullable=False),
        Column('unique_datetime', DateTime, nullable=False),
        Column('bucket_unique_datetime', DateTime, nullable=False),
        Column('same_string', String, nullable=False),
        Column('unique_string', String, nullable=False),
        Column('bucket_unique_string', String, nullable=False),
        Column('same_float', Float, nullable=False),
        Column('unique_float', Float, nullable=False),
        Column('bucket_unique_float', Float, nullable=False),
        Column('same_int', Integer, nullable=False),
        Column('unique_int', Integer, nullable=False),
        Column('bucket_unique_int', Integer, nullable=False),
    )
    metadata.create_all(engine)
    values = []
    for bucket in range(1, 20):
        for row in range(1, 10):
            values.append((
                datetime(2021, 5, 5, 11),
                datetime(2021, 5, bucket, row),
                datetime(2021, 5, 5, row),
                'hi',
                f'{bucket}-{row}',
                f'{row}',
                1.1,
                bucket * 1.0 + (row / 100.0),
                row * 1.0,
                1,
                bucket * 1000 + row,
                row
            ))
    conn.execute(sensor_readings.insert(), [
        dict(zip(
            ('same_datetime', 'unique_datetime', 'bucket_unique_datetime',
             'same_string', 'unique_string', 'bucket_unique_string',
             'same_float', 'unique_float', 'bucket_unique_float',
             'same_int', 'unique_int', 'bucket_unique_int'), value))
        for value in values
    ])
    return engine



import sqlalchemy

from collections import defaultdict
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Tuple
from typing import Set

from datools.models import Column
from datools.models import Table


MAX_SET_VALUES_TO_INCLUDE = 100
RANGE_BUCKETS = 3
RANGE_VALUED_TYPES = {
    sqlalchemy.sql.sqltypes.BIGINT,
    sqlalchemy.sql.sqltypes.DATE,
    sqlalchemy.sql.sqltypes.DATETIME,
    sqlalchemy.sql.sqltypes.DECIMAL,
    sqlalchemy.sql.sqltypes.FLOAT,
    sqlalchemy.sql.sqltypes.INTEGER,
    sqlalchemy.sql.sqltypes.NUMERIC,
    sqlalchemy.sql.sqltypes.REAL,
    sqlalchemy.sql.sqltypes.SMALLINT,
    sqlalchemy.sql.sqltypes.TIME,
    sqlalchemy.sql.sqltypes.TIMESTAMP,
}
SET_VALUED_TYPES = {
    sqlalchemy.sql.sqltypes.BOOLEAN,
    sqlalchemy.sql.sqltypes.CHAR,
    sqlalchemy.sql.sqltypes.INTEGER,
    sqlalchemy.sql.sqltypes.NCHAR,
    sqlalchemy.sql.sqltypes.NVARCHAR,
    sqlalchemy.sql.sqltypes.SMALLINT,
    sqlalchemy.sql.sqltypes.TEXT,
    sqlalchemy.sql.sqltypes.VARCHAR,
}


class ColumnStatistics:
    pass


@dataclass
class SetValuedStatistics(ColumnStatistics):
    distinct_values: int
    popular_values: list

    def __repr__(self):
        return (
            f'SetValuedStatistics(distinct_values: {self.distinct_values}, '
            f'popular_values: {self.popular_values})')

    def __eq__(self, other):
        return (self.distinct_values == other.distinct_values
                and set(self.popular_values) == set(other.popular_values))


@dataclass
class RangeValuedStatistics(ColumnStatistics):
    bucket_minimums: List[Any]

    def __repr__(self):
        return (
            f'RangeValuedStatistics(bucket_minimums: '
            f'{[minimum for minimum in self.bucket_minimums]})')


def _set_valued_statistics(
        conn: sqlalchemy.engine.Connection,
        table: Table,
        columns: List[Column]
) -> Generator[Tuple[Column, SetValuedStatistics], None, None]:
    clauses: list[str] = []
    for column in columns:
        clauses.append(
            f'COUNT(DISTINCT {column.name})'
            f'AS {column.name}_distinct_values')
    results = list(conn.execute(
        f'SELECT {", ".join(clauses)} FROM {table.name}'))[0]
    for column in columns:
        values = [
            row[column.name] for row in
            conn.execute(
                f'SELECT {column.name}, COUNT(*) AS num_rows '
                f'FROM {table.name} '
                f'GROUP BY {column.name} '
                f'ORDER BY num_rows DESC, {column.name} '
                f'LIMIT {MAX_SET_VALUES_TO_INCLUDE}')]
        yield (
            column,
            SetValuedStatistics(
                results[f'{column.name}_distinct_values'],
                values))


def _range_valued_statistics(
        conn: sqlalchemy.engine.Connection,
        table: Table,
        columns: List[Column]
) -> Generator[Tuple[Column, SetValuedStatistics], None, None]:
    bucket_clauses: list[str] = []
    first_clauses: list[str] = []
    for column in columns:
        bucket_clauses.append(
            f'{column.name}, '
            f'NTILE({RANGE_BUCKETS}) OVER (ORDER BY {column.name} ASC) '
            f'AS {column.name}_bucket')
        first_clauses.append(
            f'first_value({column.name}) OVER '
            f'(PARTITION BY {column.name}_bucket '
            f' ORDER BY {column.name} ASC '
            f' RANGE BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING) '
            f'AS {column.name}_bucket_value')
    results = conn.execute(
        f'WITH buckets AS ( '
        f'SELECT {", ".join(bucket_clauses)} '
        f'FROM {table.name} '
        f')'
        f'SELECT {", ".join(first_clauses)} '
        f'FROM buckets ')
    # The output of the window function is the number of rows in the
    # input rather than the number of range buckets. Because we have a
    # single query across multiple columns, we can't `GROUP BY
    # column_name_bucket` and get the `MIN` NTILE value for each
    # bucket. Since sqlite doesn't support CUBE/ROLLUP/GROUPING SETS,
    # we do this grouping manually by creating a `set` of the
    # `first_value` for each bucket.
    column_values: Dict[Column, List[Any]] = defaultdict(set)
    for row in results:
        for column in columns:
            column_values[column].add(row[f'{column.name}_bucket_value'])
    for column in columns:
        yield (
            column,
            RangeValuedStatistics(sorted(column_values[column])))


def column_statistics(
        engine: sqlalchemy.engine.Engine,
        table: Table,
        columns_to_ignore: Set[Column]
) -> Dict[Column, List[ColumnStatistics]]:
    metadata = sqlalchemy.MetaData()
    table_metadata = sqlalchemy.Table(
        table.name, metadata, autoload_with=engine)
    candidate_columns = [column for column in table_metadata.columns
                         if Column(column.name) not in columns_to_ignore]
    statistics: Dict[Column, List[ColumnStatistics]] = defaultdict(list)
    conn = engine.connect()
    for column, statistic in _set_valued_statistics(
            conn, table,
            [Column(column.name) for column in candidate_columns if
             type(column.type) in SET_VALUED_TYPES]):
        statistics[column].append(statistic)
    for column, statistic in _range_valued_statistics(
            conn, table,
            [Column(column.name) for column in candidate_columns if
             type(column.type) in RANGE_VALUED_TYPES]):
        statistics[column].append(statistic)
    return statistics



#!/usr/bin/env python

from datools.models import Column
from datools.models import Explanation
from datools.models import Operator
from datools.models import Predicate
from datools.explanation.algorithms import diff
from .fixtures import generate_scorpion_testdb


def test_diff():
    engine = generate_scorpion_testdb()
    candidates = diff(
        engine,
        'SELECT * FROM sensor_readings WHERE temperature > 50',
        'SELECT * FROM sensor_readings WHERE temperature <= 50',
        (Column('created_at'), Column('sensor_id'), Column('voltage'),
         Column('humidity')),
        .05,
        2.0,
        1)
    assert(candidates == [
        Explanation((Predicate(Column('voltage'), Operator.EQUALS, 2.3), ),
                    risk_ratio=9.0),
        Explanation((Predicate(Column('sensor_id'), Operator.EQUALS, '3'), ),
                    risk_ratio=5 + (1.0 / 3))])



#!/usr/bin/env python

from datools.models import Column
from datools.models import Table
from datools.table_statistics import column_statistics
from datools.table_statistics import RangeValuedStatistics
from datools.table_statistics import SetValuedStatistics
from .fixtures import generate_synthetic_testdb


def test_table_statistics():
    """Tests `column_statistics` on synthetic data in 19 buckets with 9
    values each.
    """
    engine = generate_synthetic_testdb()
    statistics = column_statistics(engine, Table('synthetic_data'), {})
    assert {
        Column('id'):
            [SetValuedStatistics(171, list(range(1, 101))),
             RangeValuedStatistics([1, 58, 115])],
        Column('same_datetime'):
            [RangeValuedStatistics([
                '2021-05-05 11:00:00.000000'])],
        Column('unique_datetime'):
            [RangeValuedStatistics([
                '2021-05-01 01:00:00.000000',
                '2021-05-07 04:00:00.000000',
                '2021-05-13 07:00:00.000000'])],
        Column('bucket_unique_datetime'):
            [RangeValuedStatistics([
                '2021-05-05 01:00:00.000000',
                '2021-05-05 04:00:00.000000',
                '2021-05-05 07:00:00.000000'])],
        Column('same_string'):
            [SetValuedStatistics(1, ['hi'])],
        Column('unique_string'):
            [SetValuedStatistics(171,
                                 sorted([f'{bucket}-{row}'
                                         for bucket in range(1, 20)
                                         for row in range(1, 10)])[:100])],
        Column('bucket_unique_string'):
            [SetValuedStatistics(9, [str(x) for x in range(1, 10)])],
        Column('same_float'):
            [RangeValuedStatistics([1.1])],
        Column('unique_float'):
            [RangeValuedStatistics([1.01, 7.04, 13.07])],
        Column('bucket_unique_float'):
            [RangeValuedStatistics([1.0, 4.0, 7.0])],
        Column('same_int'):
            [SetValuedStatistics(1, [1]), RangeValuedStatistics([1])],
        Column('unique_int'):
            [SetValuedStatistics(171,
                                 sorted([bucket * 1000 + row
                                         for bucket in range(1, 20)
                                         for row in range(1, 10)])[:100]),
             RangeValuedStatistics([1001, 7004, 13007])],
        Column('bucket_unique_int'):
            [SetValuedStatistics(9, list(range(1, 10))),
             RangeValuedStatistics([1, 4, 7])],
    } == statistics
